@@ -14,12 +14,19 @@ pub type StepFunction<T> = unsafe extern "C" fn(
 
 pub type ContinuationFunction<T> = unsafe extern "C" fn(&mut AsyncStack, T) -> cps::Result;
 
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+enum AsyncStackAction {
+    Suspend,
+    Resume,
+    Restore,
+}
+
 #[repr(C)]
 #[derive(Debug)]
 pub struct AsyncStack {
     stack: Stack,
     context: Option<*mut Context<'static>>,
-    suspended: bool,
+    next_action: AsyncStackAction,
 }
 
 impl AsyncStack {
@@ -27,7 +34,7 @@ impl AsyncStack {
         Self {
             stack: Stack::new(capacity),
             context: None,
-            suspended: false,
+            next_action: AsyncStackAction::Suspend,
         }
     }
 
@@ -46,22 +53,38 @@ impl AsyncStack {
         continuation: ContinuationFunction<T>,
         future: impl Future + Unpin,
     ) {
+        if self.next_action != AsyncStackAction::Suspend {
+            panic!("continuation not suspendable");
+        }
+
+        self.next_action = AsyncStackAction::Resume;
+
         self.stack.push(future);
         self.stack.push(step);
         self.stack.push(continuation);
-
-        self.suspended = true;
     }
 
     pub fn resume<T>(&mut self) -> (StepFunction<T>, ContinuationFunction<T>) {
+        if self.next_action != AsyncStackAction::Resume {
+            panic!("continuation not resumable");
+        }
+
+        self.next_action = AsyncStackAction::Restore;
+
         let continuation = self.pop();
         let step = self.pop();
 
         (step, continuation)
     }
 
-    pub fn restore<F: Future + Unpin>(&mut self) -> F {
-        self.pop()
+    pub fn restore<F: Future + Unpin>(&mut self) -> Option<F> {
+        if self.next_action != AsyncStackAction::Restore {
+            panic!("continuation not restorable");
+        }
+
+        self.next_action = AsyncStackAction::Suspend;
+
+        Some(self.pop())
     }
 }
 
@@ -163,6 +186,22 @@ mod tests {
         stack.set_context(&mut context);
         stack.suspend(step, continuation, future);
         stack.resume::<()>();
-        stack.restore::<TestFuture>().await;
+        stack.restore::<TestFuture>().unwrap().await;
+    }
+
+    #[tokio::test]
+    #[should_panic]
+    async fn panic_on_restore_before_resume() {
+        let waker = create_waker();
+        let mut stack = AsyncStack::new(TEST_CAPACITY);
+        let mut context = Context::from_waker(&waker);
+
+        type TestFuture = Ready<()>;
+
+        let future: TestFuture = ready(());
+
+        stack.set_context(&mut context);
+        stack.suspend(step, continuation, future);
+        stack.restore::<TestFuture>().unwrap().await;
     }
 }
