@@ -7,7 +7,7 @@ pub struct None {
 }
 
 impl None {
-    pub fn new() -> Self {
+    pub const fn new() -> Self {
         Self { _private: [] }
     }
 }
